import aiomysql
from aiomysql import InterfaceError

from .base import (
    CommandParams,
    ClientClass,
    Result,
)


class MysqlClient(ClientClass):
    ENGINE = 'MySQL'

    def __init__(self, host, username, password, dbname, port='3306'):
        super().__init__(host, username, password, dbname, port)
        self.cache = {}
        if not port:
            self.port = '3306'

    async def get_suggestions(self):
        if 'tables' not in self.cache:
            self.cache['tables'] = [list(x.values())[0] for x in (await self.get_tables()).data]

        if 'databases' not in self.cache:
            self.cache['databases'] = [list(x.values())[0] for x in (await self.get_databases()).data]

        suggestions = [f"{x} (COMMAND)" for x in self.SQL_COMMANDS]
        tables = [f"{x} (TABLE)" for x in self.cache['tables']]
        databases = [f"{x} (DATABASE)" for x in self.cache['databases']]

        return suggestions + tables + databases

    async def connect(self):
        self.connection = await aiomysql.connect(
            host=self.host,
            port=int(self.port),
            user=self.username,
            password=self.password,
            db=self.dbname,
            autocommit=True
        )

    async def change_database(self, database: str):
        self.connection = None
        self.cache.pop('tables', None)
        return await super().change_database(database)

    async def get_tables(self) -> Result:
        return await self.execute('SHOW TABLES')

    async def get_databases(self) -> Result:
        return await self.execute('SHOW DATABASES')

    async def command_use(self, command: CommandParams):
        self.cache.pop('tables', None)
        return await self.change_database(command.params)

    async def command_tables(self, command: CommandParams):
        return await self.get_tables()

    async def command_databases(self, command: CommandParams):
        return await self.get_databases()

    async def command_schema(self, command: CommandParams):
        table = command.params
        return await self.execute('SHOW CREATE TABLE %s' % table)

    async def execute(self, sql) -> Result:
        result = await self.if_command_process(sql)

        if result:
            return result

        for tries in range(2):
            try:

                if self.connection is None:
                    await self.connect()

                async with self.connection.cursor(aiomysql.DictCursor) as cur:
                    await cur.execute(sql)
                    data = await cur.fetchall()

                    return Result(data, cur.rowcount)
            except InterfaceError as exc:
                self.connection = None

                if tries == 1:
                    raise exc



import sqlite3
import asyncio

from .base import (
    ClientClass,
    Result,
)


class Sqlite3Client(ClientClass):
    ENGINE = 'Sqlite3'

    def __init__(self, filename):
        self.cache = {}
        self.filename = filename

    async def get_suggestions(self):
        if 'tables' not in self.cache:
            self.cache['tables'] = [list(x.values())[0] for x in (await self.get_tables()).data]

        suggestions = [f"{x} (COMMAND)" for x in self.SQL_COMMANDS]
        tables = [f"{x} (TABLE)" for x in self.cache['tables']]

        return suggestions + tables

    async def get_tables(self) -> Result:
        return await self.execute("SELECT name FROM sqlite_master WHERE type='table';")

    async def get_databases(self) -> Result:
        return Result([], 0)

    def _execute_sync(self, sql) -> Result:
        conn = sqlite3.connect(self.filename)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute(sql)
        rowcount = cur.rowcount
        data = [dict(x) for x in cur.fetchall()]
        if rowcount <= 0:
            rowcount = len(data)
        conn.close()

        return Result(data, rowcount)

    async def execute(self, sql) -> Result:
        sql_stripped = sql.strip()

        if sql_stripped.startswith('.tables'):
            return await self.get_tables()

        if sql_stripped.startswith('.schema '):
            first_word = sql_stripped.split(' ')[1]
            sql = f"SELECT sql FROM sqlite_master WHERE tbl_name = '{first_word}';"

        return await asyncio.to_thread(self._execute_sync, sql)

    def get_title(self) -> str:
        return f'{self.ENGINE} {self.filename}'



import aiochclient
from aiohttp import ClientSession
from aiohttp import ClientTimeout

from .base import (
    CommandParams,
    ClientClass,
    Result,
)


class ClickhouseClient(ClientClass):
    ENGINE = 'Clickhouse'

    def __init__(self, host, username, password, dbname, port='8123'):
        super().__init__(host, username, password, dbname, port)
        self.cache = {}
        if not dbname:
            self.dbname = 'default'
        if not port:
            self.port = '8123'

    async def get_suggestions(self):
        if 'tables' not in self.cache:
            self.cache['tables'] = [list(x.values())[0] for x in (await self.get_tables()).data]

        if 'databases' not in self.cache:
            self.cache['databases'] = [list(x.values())[0] for x in (await self.get_databases()).data]

        suggestions = [f"{x} (COMMAND)" for x in self.SQL_COMMANDS]
        tables = [f"{x} (TABLE)" for x in self.cache['tables']]
        databases = [f"{x} (DATABASE)" for x in self.cache['databases']]

        return suggestions + tables + databases

    async def get_tables(self) -> Result:
        return await self._execute('SHOW TABLES')

    async def get_databases(self) -> Result:
        return await self._execute('SHOW DATABASES')

    async def command_use(self, command: CommandParams):
        self.cache.pop('tables', None)
        return await self.change_database(command.params)

    async def command_tables(self, command: CommandParams):
        return await self.get_tables()

    async def command_databases(self, command: CommandParams):
        return await self.get_databases()

    async def command_schema(self, command: CommandParams):
        return await self._execute(f'SHOW CREATE TABLE {command.params}')

    async def _execute(self, sql):
        db = self.dbname

        timeout = ClientTimeout(connect=60)

        async with ClientSession(timeout=timeout) as sess:
            client = aiochclient.ChClient(
                sess,
                url=f"http://{self.host}:{self.port}",
                database=db,
                user=self.username,
                password = self.password,
            )

            data = [dict(x) for x in await client.fetch(sql.rstrip(';'), decode=True)]

            return Result(data=data, rowcount=len(data))

    async def execute(self, sql) -> Result:
        result = await self.if_command_process(sql)

        if result:
            return result

        return await self._execute(sql)



import asyncio
import threading
import operator
import json
import sys
import curses
from functools import partial
import time
from typing import Callable
from typing import Optional
import logging
from queue import LifoQueue

import kaa
import kaa.cui.main
from kaa.ui.itemlist import itemlistmode
from kaa.ui.selectlist import filterlist
import visidata
from kaa.addon import (
    alt,
    backspace,
    command,
    ctrl,
    right,
    shift,
    setup,
)
from kaa.cui.editor import TextEditorWindow
from kaa.cui.keydef import KeyEvent
from kaa.filetype.default.defaultmode import DefaultMode
from kaa.options import build_parser
from kaa.syntax_highlight import DefaultToken

from .clients.base import Result
from .sql_tokenizer import (
    CaseInsensitiveKeywords,
    NonSqlComment,
    CommandSpan,
    make_tokenizer,
    sql_editor_themes,
)
from .clients.sqlite3 import Sqlite3Client


client = None
asyncloop_thread = None

logging.basicConfig(level=logging.WARNING)


class Task:
    def __init__(self, coro, loop):
        self.coro = coro
        self.loop = loop
        self.task = None

    async def worker(self):
        return await self.coro

    def cancel(self):
        self.loop.call_soon_threadsafe(self.task.cancel)

    def is_done(self):
        if self.task is None:
            return False

        return self.task.done()

    def result(self):
        return self.task.result()

    async def run(self):
        self.task = asyncio.create_task(self.worker())
        return self.task


class AsyncLoopThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request_queue = asyncio.Queue()
        self.result_queue = LifoQueue()
        self.current_running_task = None
        self.loop = None

    async def _run(self):
        self.loop = asyncio.get_event_loop()

        while True:
            await asyncio.sleep(0.1)

    def run(self):
        asyncio.run(self._run())

    def is_done(self):
        if not self.current_running_task:
            return True
        if self.current_running_task.done():
            self.current_running_task = None
            return True
        return False

    def submit(self, coro: asyncio.coroutines):
        task = Task(coro, self.loop)
        asyncio.run_coroutine_threadsafe(task.run(), loop=self.loop)
        return task


def show_listdlg(title, candidates, callback, query=''):
    doc = filterlist.FilterListInputDlgMode.build(
        title, callback)
    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = filterlist.FilterListMode.build()
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    filterlistdoc.mode.set_candidates(candidates)
    filterlistdoc.mode.set_query(list, query.strip())
    dlg.on_console_resized()

    return dlg


def get_sel(wnd: TextEditorWindow) -> str:
    selection = wnd.screen.selection
    if not selection.is_selected():
        return
    if not selection.is_rectangular():
        selected_from, selected_to = selection.get_selrange()
        return wnd.document.gettext(selected_from, selected_to)
    data = []
    position_from, position_to, column_from, column_to = selection.get_rect_range()

    while position_from < position_to:
        position_and_col_string = selection.get_col_string(position_from, column_from, column_to)
        if position_and_col_string:
            *_, col_string = position_and_col_string
            data.append(col_string.rstrip('\n'))
        else:
            data.append('')
        position_from = wnd.document.geteol(position_from)

    return '\n'.join(data)


def get_current_sql_rows_pos(wnd: TextEditorWindow) -> list[int]:
    """Analze current editor position to find rows of sql query under the cursor"""
    pos = wnd.cursor.pos
    start_pos = pos
    rows = set()
    back_only = False

    if (
        pos < len(wnd.document.buf) and
        (
            isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), CommandSpan) or
            (
                pos > 0 and
                wnd.document.buf[pos] == '\n' and
                isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos - 1), CommandSpan)
            )
        )
    ):
        return [wnd.document.gettol(pos)]

    if (
        pos < len(wnd.document.buf) and
        (
            isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), NonSqlComment) or
            (
                pos > 0 and
                wnd.document.buf[pos] == '\n' and
                isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos - 1), NonSqlComment)
            )
        )
    ):
        return []

    for pos in range(pos, 0, -1):
        if pos >= len(wnd.document.buf):
            continue

        if (
            wnd.document.buf[pos] == ';' and
            isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), DefaultToken) and
            wnd.document.gettol(pos) == wnd.document.gettol(start_pos)
        ):
            rows.add(wnd.document.gettol(pos))
            back_only = True

        if (
            (
                (
                    (
                        wnd.document.buf[pos - 1] == ';' and
                        wnd.document.gettol(pos) != wnd.document.gettol(start_pos)
                    ) or
                    (wnd.document.buf[pos] == '\n' and (pos - 1) <= 0) or
                    (wnd.document.buf[pos] == '\n' and wnd.document.buf[pos - 1] == '\n')
                ) and
                isinstance(
                    wnd.document.mode.tokenizer.get_token_at(wnd.document, pos - 1),
                    (DefaultToken, CaseInsensitiveKeywords)
                )
            ) or
            isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), (NonSqlComment, CommandSpan)) or
            (
                wnd.document.buf[pos] == '\n' and
                isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos - 1), (NonSqlComment, CommandSpan))
            )
        ):
            break

        rows.add(wnd.document.gettol(pos))

    if back_only:
        return list(sorted(rows))

    pos = start_pos

    for pos in range(pos, len(wnd.document.buf)):
        if (
            isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), (NonSqlComment, CommandSpan)) or
            (
                (
                    wnd.document.buf[pos] == ';' or
                    (wnd.document.buf[pos] == '\n' and len(wnd.document.buf) <= pos + 1) or
                    (wnd.document.buf[pos] == '\n' and wnd.document.buf[pos + 1] == '\n')
                ) and
                isinstance(wnd.document.mode.tokenizer.get_token_at(wnd.document, pos), DefaultToken)
            )
        ):
            break

        rows.add(wnd.document.gettol(pos))

    return list(sorted(rows))


def get_expression_under_cursor(wnd: TextEditorWindow) -> str:
    line = ''
    for row in get_current_sql_rows_pos(wnd):
        _, sel = wnd.screen.document.getline(row)
        if sel:
            line += sel

    return line


def print_center(window: curses.window, text: str):
    num_rows, num_cols = window.getmaxyx()
    x = num_cols // 2 - len(text) // 2
    y = num_rows // 2
    window.addstr(y, x, text)
    window.refresh()


def await_and_print_time(
        wnd: TextEditorWindow,
        coro: asyncio.coroutines
) -> Result:
    start = time.time()

    posy, _ = wnd.get_cursor_loc()

    win = curses.newwin(3, 50, max(posy - 3, 0), 5)

    task = asyncloop_thread.submit(coro)

    while time.time() - start < 0.3:
        time.sleep(0.1)
        if task.is_done():
            return task.result()

    try:
        win.box()

        while not task.is_done():
            wnd.mainframe._cwnd.timeout(0)
            key = wnd.mainframe._cwnd.getch()

            if key == 27:
                task.cancel()

            print_center(win, f'Running (press ESC to cancel): {round(time.time() - start, 2)}s'.ljust(45, ' '))

            time.sleep(0.1)
    finally:
        del win
    return task.result()


def fix_visidata_curses():
    if visidata.color.colors.color_pairs:
        for (fg, bg), (pairnum, _) in visidata.color.colors.color_pairs.items():
            curses.init_pair(pairnum, fg, bg)


def fix_kaa_curses(wnd: TextEditorWindow):
    curses.endwin()
    kaa.app.show_cursor(1)

    for pairnum, (fg, bg) in enumerate(kaa.app.colors.pairs.keys()):
        curses.init_pair(pairnum, fg, bg)

    wnd.draw_screen(force=True)


def predictions_weights(query, candidate):
    if query == candidate:
        return (0, candidate)
    if candidate.startswith(query):
        return (1, candidate)
    return (2, candidate)


def run_corutine_and_show_result(wnd: TextEditorWindow, coro: asyncio.coroutines):
    start = time.time()
    end = None
    message = ''

    try:
        try:
            result = await_and_print_time(wnd, coro)
        except asyncio.CancelledError:
            end = time.time()
            message = 'Cancelled'
            return

        end = time.time()
        message = str(result)

        if not result or not result.data:
            return

        fix_visidata_curses()

        visidata.vd.run()
        visidata.vd.view(result.data)
    except Exception as exc:
        end = time.time()
        message = str(exc)
    finally:
        wnd.document.set_title(client.get_title())
        kaa.app.messagebar.set_message(f'{round(end - start, 2)}s {message}')
        fix_kaa_curses(wnd)


@command('db.query')
def db_query(wnd: TextEditorWindow):
    sel = get_sel(wnd)

    if not sel:
        sel = get_expression_under_cursor(wnd)

    if not sel or not sel.strip():
        kaa.app.messagebar.set_message("Nothing to execute")
        return

    selection = sel.strip()
    run_corutine_and_show_result(wnd, client.execute(selection))


@command('db.show_prediction')
def show_prediction(wnd: TextEditorWindow):
    from_pos = None
    to_pos = None
    word = ''

    res = wnd.document.mode.get_word_at(wnd.cursor.pos)

    if res:
        from_pos, to_pos, _ = res
        if from_pos != to_pos:
            word = wnd.document.gettext(from_pos, to_pos)

    candidates = await_and_print_time(wnd, client.get_suggestions())
    sorted_candidates = sorted(candidates, key=partial(predictions_weights, word))

    def callback(result):
        if not result:
            return

        type_idx = operator.indexOf(reversed(result), '(') + 2
        result = result[:-type_idx]

        pos = wnd.cursor.pos

        if from_pos is not None and from_pos < pos:
            wnd.cursor.left(word=True)

            wnd.document.mode.delete_string(wnd, from_pos, to_pos)

        wnd.document.mode.put_string(wnd, result)
        return

    show_listdlg('Select', sorted_candidates, callback, query=word)


@command('db.show_tables')
def db_show_tables(wnd: TextEditorWindow):
    run_corutine_and_show_result(wnd, client.get_tables())


@command('db.show_databases')
def db_show_databases(wnd: TextEditorWindow):
    run_corutine_and_show_result(wnd, client.get_databases())


def on_keypressed(
        self: DefaultMode,
        original_fn: Callable,
        wnd: TextEditorWindow,
        event: KeyEvent,
        key: Optional[str],
        commands: list[str],
        candidate: list[tuple]
):
    pos = wnd.cursor.pos
    tol = wnd.document.gettol(pos)
    wnd.document.marks['current_script'] = (0, tol)
    wnd.document.style_updated()
    wnd.document.set_title(client.get_title())
    return original_fn(wnd, event, key, commands, candidate)


def on_cursor_located(
        self: DefaultMode,
        original_fn: Callable,
        wnd: TextEditorWindow,
        *args, **kwargs
):
    wnd.document.highlights = []
    for id, row_pos in enumerate(get_current_sql_rows_pos(wnd)):
        wnd.document.highlights.append(
            row_pos
        )
    return original_fn(wnd, *args, **kwargs)


def get_line_overlays(self: DefaultMode, original_fn: Callable) -> dict[int, str]:
    highlights = {}
    highlights.update(original_fn())

    for pos in self.document.highlights:
        highlights[pos] = 'cursor-row'

    return highlights


@setup('kaa.filetype.default.defaultmode.DefaultMode')
def editor(mode: DefaultMode):
    # register command to the mode
    mode.add_command(db_query)
    mode.add_command(db_show_tables)
    mode.add_command(db_show_databases)
    mode.on_keypressed = partial(on_keypressed, mode, mode.on_keypressed)
    # To determine sql expression under current cursor after each key press
    mode.on_cursor_located = partial(on_cursor_located, mode, mode.on_cursor_located)
    # To highligt lines determined in on_cursor_located
    mode.get_line_overlays = partial(get_line_overlays, mode, mode.get_line_overlays)


    # add key bind th execute 'run.query'
    mode.add_keybinds(keys={
        (alt, '1'): 'db.show_prediction',
        (alt, 'r'): 'db.query',
        (alt, 't'): 'db.show_tables',
        (alt, 'e'): 'db.show_databases',
        (ctrl, 's'): 'file.save',
        (ctrl, 'f'): 'search.showsearch',
        (ctrl, 'r'): 'search.showreplace',
        (ctrl, 'q'): 'file.quit',
        (alt, backspace): 'edit.backspace.word'
    })

    mode.SHOW_LINENO = True
    # Syntax highlight
    mode.tokenizer = make_tokenizer()
    mode.themes.append(sql_editor_themes)


def main():
    global client
    global asyncloop_thread

    args_parser = build_parser()

    args_parser.description = 'DB connection tool'
    args_parser.add_argument('--config', '-c', dest='config', help='specify config path', default='')
    args_parser.add_argument('--host', '-H', dest='host', help='specify host name', default='')
    args_parser.add_argument('--user', '-u', dest='user', help='specify user name', required=False)
    args_parser.add_argument('--password', '-p', dest='password', default='', help='specify raw password')
    args_parser.add_argument('--port', '-P', dest='port', default='', help='specify port')
    args_parser.add_argument('--engine', '-E', dest='engine', help='specify db engine', required=False,
        choices=['clickhouse', 'mysql', 'postgres', 'sqlite3'])
    args_parser.add_argument('--dbname', '-d', dest='dbname', help='specify db name', required=False)
    args_parser.add_argument('--filepath', '-f', dest='filepath', help='specify db filepath', required=False)

    args = args_parser.parse_args()

    host = args.host
    username = args.user
    password = ''

    if args.password:
        password = args.password

    port = args.port
    engine = args.engine
    dbname = args.dbname
    filepath = args.filepath

    if args.config:
        with open(args.config) as f:
            config = json.load(f)

        if not host or host == '127.0.0.1':
            host = config.get('host', '')
        if not port:
            port = config.get('port', '')
        if not username:
            username = config.get('username', '')
        if not password:
            password = config.get('password', '')
        if not dbname:
            dbname = config.get('dbname', '')
        if not engine:
            engine = config.get('engine', '')
        if not filepath:
            filepath = config.get('filepath', '')

    # imported here to make db libs dependencies optional
    if engine == 'clickhouse':
        from .clients.clickhouse import ClickhouseClient
        client = ClickhouseClient(host, username, password, dbname, port=port)
    if engine == 'mysql':
        from .clients.mysql import MysqlClient
        client = MysqlClient(host, username, password, dbname, port=port)
    if engine == 'postgres':
        from .clients.postgres import PostgresClient
        client = PostgresClient(host, username, password, dbname, port=port)
    if engine == 'sqlite3':
        client = Sqlite3Client(filepath)

    if not client:
        args_parser.print_help(sys.stderr)
        print('Invalid engine specified')
        sys.exit(1)

    asyncloop_thread = AsyncLoopThread(daemon=True)
    asyncloop_thread.start()

    kaa.cui.main.opt = args
    kaa.cui.main._init_term()
    curses.wrapper(kaa.cui.main.main)


if __name__ == '__main__':
    main()



from kaa import doc_re
from kaa.syntax_highlight import (
    Keywords,
    SingleToken,
    Span,
    Token,
    Tokenizer,
)
from kaa.theme import Style


KEYWORDS = [
    'SELECT', 'UPDATE', 'DELETE', 'DROP', 'ALTER', 'COLUMN', 'USE',
    'FROM', 'JOIN', 'OUTER', 'INNER', 'LIMIT', 'ORDER BY', 'AS',
    'SHOW', 'FROM', 'WHERE', 'DESC', 'TABLES', 'CREATE', 'TABLE',
    'SET', 'IS', 'NOT', 'NULL', 'ON', 'IN', 'LIKE', 'ILIKE', 'AND',
    'OR', 'INSERT', 'INTO', 'VALUES', 'INTERVAL', 'GROUP', 'BY',
    'HAVING', 'GRANT', 'LEFT', 'RIGHT', 'FULL', 'CROSS'
]


sql_editor_themes = {
    'basic': [
        Style('string', 'Green', None, bold=True),
        Style('number', 'Yellow', None, bold=True),
    ]
}

class CaseInsensitiveKeywords(Keywords):
    def re_start(self) -> str:
        tokens = []
        for token in self._tokens:
            new_token = ''
            token = doc_re.escape(token)
            for char in token:
                if char.lower() != char.upper():
                    new_token += '[' + char.lower() + char.upper() + ']'
                    continue
                new_token = new_token + char
            if new_token:
                tokens.append(new_token)

        return rf'\b({"|".join(tokens)})\b'


class NonSqlComment(Span):
    pass


class CommandSpan(Span):
    pass

def sqleditor_tokens() -> list[tuple[str, Token]]:
    return [
        ("directive", CommandSpan('directive', r'^(\s+)?\.', '(\s|;|$)')),
        ('comment1', Span('comment', r'-- ', '$')),
        ('comment2', NonSqlComment('comment', r'\#', '$')),
        ("string1", Span('string', '"', '"', escape='\\')),
        ("string2", Span('string', "'", "'", escape='\\')),
        ("number", SingleToken('number', [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])),
        ("keyword", CaseInsensitiveKeywords('keyword', KEYWORDS)),
    ]


def make_tokenizer() -> Tokenizer:
    return Tokenizer(tokens=sqleditor_tokens())



import re
import aiopg
from psycopg2 import InterfaceError
from psycopg2.extras import RealDictCursor

from .base import (
    CommandParams,
    ClientClass,
    Result,
)


SHOW_CR_TAB_RE = re.compile('SHOW\s+CREATE\s+TABLE\s+(.*)$', re.IGNORECASE)


class PostgresClient(ClientClass):
    ENGINE = 'PostgreSQL'

    def __init__(self, host, username, password, dbname, port='5432'):
        self.cache = {}
        super().__init__(host, username, password, dbname, port)
        if not port:
            self.port = '5432'

    async def get_suggestions(self):
        if 'tables' not in self.cache:
            self.cache['tables'] = [list(x.values())[0] for x in (await self.get_tables()).data]

        if 'databases' not in self.cache:
            self.cache['databases'] = [list(x.values())[0] for x in (await self.get_databases()).data]

        suggestions = [f"{x} (COMMAND)" for x in self.SQL_COMMANDS]
        tables = [f"{x} (TABLE)" for x in self.cache['tables']]
        databases = [f"{x} (DATABASE)" for x in self.cache['databases']]

        return suggestions + tables + databases

    async def connect(self):
        self.connection = await aiopg.connect(
            host=self.host,
            port=int(self.port),
            user=self.username,
            password=self.password,
            dbname=self.dbname,
        )

    async def change_database(self, database: str):
        self.cache.pop('tables', None)
        if self.connection:
            await self.connection.close()
        self.connection = None
        return await super().change_database(database)

    async def get_tables(self) -> Result:
        sql = (
            "SELECT table_name FROM information_schema.tables "
            "WHERE table_schema='public' AND table_type='BASE TABLE';"
        )
        return await self.execute(sql)

    async def get_databases(self) -> Result:
        sql = "SELECT datname FROM pg_database;"
        return await self.execute(sql)

    async def show_create_table(self, table_name: str) -> Result:
        # Columns
        result = await self.execute(f"""
            SELECT
                a.attname AS column_name,
                pg_catalog.format_type(a.atttypid, a.atttypmod) AS data_type,
                CASE
                    WHEN a.attnotnull THEN ' NOT NULL'
                    ELSE ''
                END AS not_null,
                COALESCE(pg_catalog.pg_get_expr(ad.adbin, ad.adrelid), '') AS default_value
            FROM
                pg_catalog.pg_attribute a
            LEFT JOIN
                pg_catalog.pg_attrdef ad ON a.attrelid = ad.adrelid AND a.attnum = ad.adnum
            WHERE
                a.attrelid = '{table_name}'::regclass AND a.attnum > 0 AND NOT a.attisdropped
            ORDER BY
                a.attnum;
        """)

        columns = result.data

        # Constraints
        result = await self.execute(f"""
            SELECT
                pg_catalog.pg_get_constraintdef(con.oid, true) as condef
            FROM
                pg_catalog.pg_constraint con
            WHERE
                con.conrelid = '{table_name}'::regclass;
        """)

        constraints = result.data

        # Partitioning
        result = await self.execute(f"""
            SELECT
                partstrat,
                 pg_catalog.pg_get_partkeydef(pt.partrelid) as partition_key
            FROM
                pg_catalog.pg_partitioned_table pt
            WHERE
                pt.partrelid = '{table_name}'::regclass;
        """)
        partition_info = result.data[0] if result.data else None

        # Partitions
        result = await self.execute(f"""
            SELECT
                c.relname AS partition_name,
                pg_get_expr(c.relpartbound, c.oid) AS partition_expr
            FROM
                pg_class c
            JOIN
                pg_inherits i ON c.oid = i.inhrelid
            WHERE
                i.inhparent = '{table_name}'::regclass
            ORDER BY
                c.relname;
        """)

        partitions = result.data

        # Indexes
        result = await self.execute(f"""
            SELECT
                indexname,
                indexdef
            FROM
                pg_catalog.pg_indexes
            WHERE
                tablename = '{table_name.split('.')[-1]}';
        """)

        indexes = result.data

        # Child tables
        result = await self.execute(f"""
            SELECT c.relname AS child_table
            FROM pg_inherits
            JOIN pg_class c ON pg_inherits.inhrelid = c.oid
            JOIN pg_class p ON pg_inherits.inhparent = p.oid
            WHERE p.relname = '{table_name.split('.')[-1]}'
        """)

        child_tables = result.data

        create_table_query = f"-- approximate table schema\nCREATE TABLE {table_name} (\n"
        column_definitions = []

        for column in columns:
            column = list(column.values())

            column_definition = f"    {column[0]} {column[1]}{column[2]}"
            if column[3]:
                column_definition += f" DEFAULT {column[3]}"
            column_definitions.append(column_definition)

        create_table_query += ",\n".join(column_definitions)

        for constraint in constraints:
            constraint = list(constraint.values())
            create_table_query += f",\n    {constraint[0]}"

        create_table_query += "\n)"

        if partition_info:
            part_method, part_key = list(partition_info.values())
            part_method = part_method.upper()
            if part_method == 'R':
                part_method = 'RANGE'
            elif part_method == 'L':
                part_method = 'LIST'
            elif part_method == 'H':
                part_method = 'HASH'
            create_table_query += f"\nPARTITION BY {part_key}"

        create_table_query += ";"

        for index in indexes:
            index = list(index.values())
            create_table_query += f"\n{index[1]};"

        child_tables = [list(row.values())[0] for row in child_tables]

        if child_tables:
            create_table_query += f"\n-- Child tables: {', '.join(child_tables)}"

        if partitions:
            for partition in partitions:
                partition = list(partition.values())

                # Child tables
                if not partition[1]:
                    continue

                create_table_query += f"\n\nCREATE TABLE {partition[0]} PARTITION OF {table_name}\n    {partition[1]};"

        return Result(data=[{'res': create_table_query}], rowcount=1)

    async def command_use(self, command: CommandParams):
        self.cache.pop('tables', None)
        return await self.change_database(command.params)

    async def command_tables(self, command: CommandParams):
        return await self.get_tables()

    async def command_databases(self, command: CommandParams):
        return await self.get_databases()

    async def command_schema(self, command: CommandParams):
        table_name = command.params
        return await self.show_create_table(table_name)

    async def execute(self, sql) -> Result:
        result = await self.if_command_process(sql)

        if result:
            return result

        for tries in range(2):
            try:
                if self.connection is None:
                    await self.connect()

                async with self.connection.cursor(cursor_factory=RealDictCursor) as cur:
                    await cur.execute(sql)
                    rowcount = cur.rowcount
                    result = Result(rowcount=rowcount)

                    result.data = await cur.fetchall()

                    return result
            except InterfaceError as exc:
                self.connection = None

                if tries == 1:
                    raise exc



import os

from setuptools import (
    find_packages,
    setup,
)


def get_requirements():
    basedir = os.path.dirname(__file__)
    try:
        with open(os.path.join(basedir, 'requirements.txt')) as f:
            return f.readlines()
    except FileNotFoundError:
        print(os.listdir(basedir))
        print(os.path.join(basedir, 'requirements.txt'))
        raise RuntimeError('No requirements info found.')


setup(
    name='dbcls',
    version='1.1.8',
    license='BSD',
    author='Maksim Nikitenko',
    author_email='iam@sets88.com',
    packages=find_packages(),
    description='dbcls is a versatile client that supports various databases',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=get_requirements(),
    python_requires='>=3.9',
    entry_points={
        'console_scripts': [
            'dbcls = dbcls:main',
        ]
    }
)



import abc
import re
from dataclasses import (
    dataclass,
    field,
)


COMMAND_RE = re.compile(r'\.([a-zA-Z_0-9]+)\s*(.*)', re.IGNORECASE)


@dataclass
class CommandParams:
    command: str
    params: str


@dataclass
class Result:
    data: list[dict] = field(default_factory=list)
    rowcount: int = 0
    message: str = ''

    def __str__(self) -> str:
        if self.message:
            return self.message

        if self.data:
            return f'{self.rowcount} rows returned'

        if self.rowcount:
            return f'{self.rowcount} rows affected'

        return 'Empty set'


class ClientClass(abc.ABC):
    ENGINE = ''

    COMMANDS = [
        'tables', 'databases', 'schema', 'use'
    ]

    SQL_COMMANDS = [
        'SELECT', 'INSERT', 'UPDATE', 'DELETE', 'CREATE', 'ALTER', 'DROP', 'WHERE', 'TRUNCATE', 'USE', 'SHOW', 'DESCRIBE',
        'EXPLAIN', 'DESC', 'RENAME', 'GRANT', 'REVOKE', 'SET', 'BEGIN', 'COMMIT', 'ROLLBACK' 'ANALYZE', 'OPTIMIZE',
        'KILL', 'FROM', 'GROUP BY', 'ORDER BY', 'LIMIT', 'OFFSET', 'HAVING', 'JOIN', 'LEFT JOIN', 'RIGHT JOIN',
        'FULL JOIN', 'INNER JOIN', 'OUTER JOIN', 'CROSS JOIN', 'ON', 'AND', 'OR', 'NOT', 'IN', 'LIKE', 'BETWEEN'
    ]

    def __init__(self, host: str, username: str, password: str, dbname: str, port: str):
        self.host = host
        self.username = username
        self.password = password
        self.dbname = dbname
        self.port = port
        self.connection = None

    async def get_suggestions(self):
        return [f"{x} (COMMAND)" for x in self.SQL_COMMANDS]

    @abc.abstractmethod
    def get_databases(self) -> Result:
        pass

    @abc.abstractmethod
    def get_tables(self) -> Result:
        pass

    def get_internal_command_params(self, sql: str) -> list[str]:
        command = sql.strip().rstrip(';')
        if not command or not command.startswith('.'):
            return

        command, params = COMMAND_RE.match(command).groups()
        command = command.lower()
        if command not in self.COMMANDS:
            return

        return CommandParams(command, params)

    async def if_command_process(self, sql: str) -> Result:
        command = self.get_internal_command_params(sql)

        if not command:
            return

        if hasattr(self, f'command_{command.command}'):
            return await getattr(self, f'command_{command.command}')(command)

    async def change_database(self, database: str):
        old_db = self.dbname
        self.dbname = database
        try:
            await self.execute('SELECT 1')
            return Result(message=f'You are now connected to database "{database}"')
        except Exception:
            self.dbname = old_db
            raise

    def get_title(self) -> str:
        return f'{self.ENGINE} {self.host}:{self.port} {self.dbname}'

    @abc.abstractmethod
    async def execute(self, sql) -> Result:
        pass
